@@ -20,7 +20,7 @@ from buidl.helper import (
 class FieldElement:
     def __init__(self, num, prime):
         if num >= prime or num < 0:
-            error = "Num {} not in field range 0 to {}".format(num, prime - 1)
+            error = f"Num {num} not in field range 0 to {prime - 1}"
             raise ValueError(error)
         self.num = num
         self.prime = prime
@@ -35,7 +35,7 @@ class FieldElement:
         return not (self == other)
 
     def __repr__(self):
-        return "FieldElement_{}({})".format(self.prime, self.num)
+        return f"FieldElement_{self.prime}({self.num})"
 
     def __add__(self, other):
         if self.prime != other.prime:
@@ -113,7 +113,7 @@ class Point:
         # y**2 == x**3 + a*x + b
         if self.y ** 2 != self.x ** 3 + a * x + b:
             # if not, raise a ValueError
-            raise ValueError("({}, {}) is not on the curve".format(self.x, self.y))
+            raise ValueError(f"({self.x}, {self.y}) is not on the curve")
 
     def __eq__(self, other):
         return (
@@ -131,13 +131,11 @@ class Point:
         if self.x is None:
             return "Point(infinity)"
         else:
-            return "Point({},{})_{}".format(self.x.num, self.y.num, self.x.prime)
+            return f"Point({self.x.num},{self.y.num})_{self.x.prime}"
 
     def __add__(self, other):
         if self.a != other.a or self.b != other.b:
-            raise TypeError(
-                "Points {}, {} are not on the same curve".format(self, other)
-            )
+            raise TypeError(f"Points {self}, {other} are not on the same curve")
         # Case 0.0: self is the point at infinity, return other
         if self.x is None:
             return other
@@ -234,7 +232,7 @@ class S256Point(Point):
         if self.x is None:
             return "S256Point(infinity)"
         else:
-            return "S256Point({},{})".format(hex(self.x.num), hex(self.y.num))
+            return f"S256Point({self.sec().hex()})"
 
     def __rmul__(self, coefficient):
         # we want to mod by N to make this simple
@@ -248,9 +246,6 @@ class S256Point(Point):
         else:
             return super().__add__(other)
 
-    def negate(self):
-        return -1 * self
-
     def sec(self, compressed=True):
         # returns the binary version of the sec format, NOT hex
         # if compressed, starts with b'\x02' if self.y.num is even, b'\x03' if self.y is odd
@@ -317,7 +312,7 @@ class S256Point(Point):
         """Returns the p2pkh address string"""
         return self.p2pkh_script(compressed).address(network)
 
-    def bech32_address(self, network="mainnet"):
+    def p2wpkh_address(self, network="mainnet"):
         """Returns the p2wpkh bech32 address string"""
         return self.p2wpkh_script().address(network)
 
@@ -352,16 +347,20 @@ class S256Point(Point):
         return self.verify(z, sig)
 
     def verify_schnorr(self, msg, schnorr_sig):
-        if self.y.num % 2 == 1:
-            point = self.negate()
+        if self.parity:
+            point = -1 * self
         else:
             point = self
         if schnorr_sig.r.x is None:
             return False
         message = schnorr_sig.r.bip340() + point.bip340() + msg
-        e = big_endian_to_int(hash_challenge(message)) % N
-        result = schnorr_sig.s * G + (N - e) * point
-        return result == schnorr_sig.r
+        challenge = big_endian_to_int(hash_challenge(message)) % N
+        result = -challenge * point + schnorr_sig.s
+        if result.x is None:
+            return False
+        if result.parity:
+            return False
+        return result.bip340() == schnorr_sig.r.bip340()
 
     @classmethod
     def parse(cls, binary):
@@ -413,6 +412,13 @@ class S256Point(Point):
             beta = S256Field(P - beta.num)
         return cls(x, beta)
 
+    @classmethod
+    def combine(cls, points):
+        sum_point = points[0]
+        for point in points[1:]:
+            sum_point += point
+        return sum_point
+
 
 G = S256Point(
     0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798,
@@ -530,18 +536,18 @@ class PrivateKey:
         return Signature(r, s)
 
     def sign_schnorr(self, msg, aux):
+        if self.point.parity:
+            d = N - self.secret
+        else:
+            d = self.secret
         if len(msg) != 32:
             raise ValueError("msg needs to be 32 bytes")
         if len(aux) != 32:
             raise ValueError("aux needs to be 32 bytes")
-        if self.point.y.num % 2 == 0:
-            d = self.secret
-        else:
-            d = N - self.secret
         t = xor_bytes(int_to_big_endian(d, 32), hash_aux(aux))
         k = big_endian_to_int(hash_nonce(t + self.point.bip340() + msg)) % N
         r = k * G
-        if r.y.num % 2 == 1:
+        if r.parity:
             k = N - k
             r = k * G
         message = r.bip340() + self.point.bip340() + msg
@@ -549,7 +555,7 @@ class PrivateKey:
         s = (k + e * d) % N
         sig = SchnorrSignature(r, s)
         if not self.point.verify_schnorr(msg, sig):
-            raise RuntimeError("bad sig")
+            raise RuntimeError("Bad Signature")
         return sig
 
     def deterministic_k(self, z):
