@@ -41,10 +41,10 @@ class Script:
                 if OP_CODE_NAMES.get(command):
                     name = OP_CODE_NAMES.get(command)
                 else:
-                    name = "OP_[{}]".format(command)
-                result += "{} ".format(name)
+                    name = f"OP_[{command}]"
+                result += f"{name} "
             else:
-                result += "{} ".format(command.hex())
+                result += f"{command.hex()} "
         return result
 
     def __eq__(self, other):
@@ -169,9 +169,9 @@ class Script:
                     if not operation(stack, altstack):
                         print("bad op: {}".format(OP_CODE_NAMES[command]))
                         return False
-                elif command in (172, 173, 174, 175, 186):
-                    # these are signing operations, they need the tx and input index
-                    # to check against
+                elif command in (172, 173, 174, 175, 177, 178, 186):
+                    # SIG ops (172-175, 186) and CLTV/CSV (177, 178)
+                    # are operations that need the tx and input index
                     if not operation(stack, tx_obj, input_index):
                         print("bad op: {}".format(OP_CODE_NAMES[command]))
                         return False
@@ -225,9 +225,7 @@ class Script:
                     witness_script = witness.items[-1]
                     if s256 != sha256(witness_script):
                         print(
-                            "bad sha256 {} vs {}".format(
-                                s256.hex(), sha256(witness_script).hex()
-                            )
+                            f"bad sha256 {s256.hex()} vs {sha256(witness_script).hex()}"
                         )
                         return False
                     # hashes match! now add the Witness Script
@@ -235,7 +233,7 @@ class Script:
                     witness_script_commands = Script.parse(stream).commands
                     commands.extend(witness_script_commands)
                 # witness program version 1 rule. if stack commands are:
-                # 0 <32 byte hash> this is p2tr
+                # 1 <32 byte hash> this is p2tr
                 elif len(stack) == 2 and stack[0] == b"\x01" and len(stack[1]) == 32:
                     if len(witness) == 0:
                         print("stack in witness v1 empty")
@@ -256,8 +254,10 @@ class Script:
                         tweak_point = control_block.tweak_point(tap_leaf)
                         # the tweak point should be what's on the stack
                         if tweak_point.parity != control_block.parity:
+                            print("bad tweak point parity")
                             return False
                         if tweak_point.bip340() != stack.pop():
+                            print("bad tweak point")
                             return False
                         # pop off the 1 and start fresh
                         stack.pop()
@@ -527,7 +527,9 @@ class P2TRScriptPubKey(ScriptPubKey):
         elif type(point) == bytes:
             raw_point = point
         else:
-            raise TypeError("To initialize P2TRScriptPubKey, a point is needed")
+            raise TypeError(
+                f"To initialize P2TRScriptPubKey, a point is needed {point}"
+            )
         self.commands = [0x51, raw_point]
 
     def address(self, network="mainnet"):
@@ -537,17 +539,10 @@ class P2TRScriptPubKey(ScriptPubKey):
         # convert to bech32 address using encode_bech32_checksum
         return encode_bech32_checksum(witness_program, network)
 
-
-class P2PKTapScript(ScriptPubKey):
-    def __init__(self, point):
-        super().__init__()
-        if type(point) == S256Point:
-            raw_point = point.bip340()
-        elif type(point) == bytes:
-            raw_point = point
-        else:
-            raise TypeError(f"To initialize P2PKTapScript, a point is needed {point}")
-        self.commands = [raw_point, 0xAC]
+    @classmethod
+    def from_address(cls, address):
+        _, _, point_raw = decode_bech32(address)
+        return cls(point_raw)
 
 
 class WitnessScript(Script):
@@ -621,4 +616,4 @@ def address_to_script_pubkey(s):
             # p2wskh
             return P2WSHScriptPubKey(decode_bech32(s)[2])
 
-    raise RuntimeError("unknown type of address: {}".format(s))
+    raise RuntimeError(f"unknown type of address: {s}")
