@@ -1,9 +1,10 @@
 from io import BytesIO
-
 from urllib.request import Request, urlopen
 
 import json
 
+from buidl.bech32 import decode_bech32
+from buidl.ecc import SchnorrSignature
 from buidl.helper import (
     big_endian_to_int,
     decode_base58,
@@ -24,18 +25,24 @@ from buidl.helper import (
 )
 from buidl.script import (
     P2PKHScriptPubKey,
+    P2SHScriptPubKey,
+    P2WPKHScriptPubKey,
+    P2WSHScriptPubKey,
+    P2TRScriptPubKey,
     RedeemScript,
     Script,
     ScriptPubKey,
     WitnessScript,
 )
+from buidl.taproot import MultiSigTapScript
+from buidl.timelock import Locktime, Sequence
 from buidl.witness import Witness
 
 
 URL = {
     "mainnet": "https://blockstream.info/api",
     "testnet": "https://blockstream.info/testnet/api",
-    "signet": "https://explorer.bc-2.jp/api",
+    "signet": "https://mempool.space/signet/api",
 }
 
 
@@ -55,7 +62,7 @@ class TxFetcher:
             try:
                 raw = bytes.fromhex(response)
             except ValueError:
-                raise ValueError("unexpected response: {}".format(response))
+                raise ValueError(f"unexpected response: {response}")
             tx = Tx.parse(BytesIO(raw), network=network)
             # make sure the tx we got matches to the hash we requested
             if tx.segwit:
@@ -63,7 +70,7 @@ class TxFetcher:
             else:
                 computed = hash256(raw)[::-1].hex()
             if computed != tx_id:
-                raise RuntimeError("server lied: {} vs {}".format(computed, tx_id))
+                raise RuntimeError(f"server lied: {computed} vs {tx_id}")
             cls.cache[tx_id] = tx
         cls.cache[tx_id].network = network
         return cls.cache[tx_id]
@@ -86,12 +93,15 @@ class Tx:
     command = b"tx"
 
     def __init__(
-        self, version, tx_ins, tx_outs, locktime, network="mainnet", segwit=False
+        self, version, tx_ins, tx_outs, locktime=None, network="mainnet", segwit=False
     ):
         self.version = version
         self.tx_ins = tx_ins
         self.tx_outs = tx_outs
-        self.locktime = locktime
+        if locktime is None:
+            self.locktime = Locktime()
+        else:
+            self.locktime = Locktime(locktime)
         self.network = network
         self.segwit = segwit
         self._hash_prevouts = None
@@ -104,19 +114,15 @@ class Tx:
         self._sha_outputs = None
 
     def __repr__(self):
-        tx_ins = ""
-        for tx_in in self.tx_ins:
-            tx_ins += tx_in.__repr__() + "\n"
-        tx_outs = ""
-        for tx_out in self.tx_outs:
-            tx_outs += tx_out.__repr__() + "\n"
-        return "tx: {}\nversion: {}\ntx_ins:\n{}\ntx_outs:\n{}\nlocktime: {}\n".format(
-            self.hash().hex(),
-            self.version,
-            tx_ins,
-            tx_outs,
-            self.locktime,
-        )
+        tx_ins = "\n".join([str(txi) for txi in self.tx_ins])
+        tx_outs = "\n".join([str(txo) for txo in self.tx_outs])
+        return f"""
+tx: {self.hash().hex()}
+version: {self.version}
+locktime: {self.locktime}
+tx_ins:\n{tx_ins}
+tx_outs:\n{tx_outs}
+"""
 
     def clone(self):
         tx_obj = self.__class__.parse(BytesIO(self.serialize()), network=self.network)
@@ -133,6 +139,12 @@ class Tx:
         """Binary hash of the legacy serialization"""
         return hash256(self.serialize_legacy())[::-1]
 
+    def vbytes(self):
+        if self.segwit:
+            return len(self.serialize()) - (len(self.serialize_witness()) + 2) * 3 // 4
+        else:
+            return len(self.serialize())
+
     @classmethod
     def parse_hex(cls, s, network="mainnet"):
         """Parses a transaction from a hex string"""
@@ -173,7 +185,7 @@ class Tx:
         for _ in range(num_outputs):
             outputs.append(TxOut.parse(s))
         # locktime is 4 bytes, little-endian
-        locktime = little_endian_to_int(s.read(4))
+        locktime = Locktime.parse(s)
         # return an instance of the class (cls(...))
         return cls(version, inputs, outputs, locktime, network=network, segwit=False)
 
@@ -186,7 +198,7 @@ class Tx:
         # next two bytes need to be 0x00 and 0x01, otherwise raise RuntimeError
         marker = s.read(2)
         if marker != b"\x00\x01":
-            raise RuntimeError("Not a segwit transaction {}".format(marker))
+            raise RuntimeError(f"Not a segwit transaction {marker}")
         # num_inputs is a varint, use read_varint(s)
         num_inputs = read_varint(s)
         # each input needs parsing, create inputs array
@@ -205,7 +217,7 @@ class Tx:
         for tx_in in inputs:
             tx_in.witness = Witness.parse(s)
         # locktime is 4 bytes, little-endian
-        locktime = little_endian_to_int(s.read(4))
+        locktime = Locktime.parse(s)
         # return an instance of the class (cls(...))
         return cls(version, inputs, outputs, locktime, network=network, segwit=True)
 
@@ -232,7 +244,15 @@ class Tx:
             # serialize each output
             result += tx_out.serialize()
         # serialize locktime (4 bytes, little endian)
-        result += int_to_little_endian(self.locktime, 4)
+        result += self.locktime.serialize()
+        return result
+
+    def serialize_witness(self):
+        result = b""
+        # add the witness data for each input
+        for tx_in in self.tx_ins:
+            # serialize the witness field
+            result += tx_in.witness.serialize()
         return result
 
     def serialize_segwit(self):
@@ -253,12 +273,9 @@ class Tx:
         for tx_out in self.tx_outs:
             # serialize each output
             result += tx_out.serialize()
-        # add the witness data for each input
-        for tx_in in self.tx_ins:
-            # serialize the witness field
-            result += tx_in.witness.serialize()
+        result += self.serialize_witness()
         # serialize locktime (4 bytes, little endian)
-        result += int_to_little_endian(self.locktime, 4)
+        result += self.locktime.serialize()
         return result
 
     def fee(self):
@@ -306,7 +323,7 @@ class Tx:
             else:
                 script_sig = None
                 if hash_type & 3 in (SIGHASH_NONE, SIGHASH_SINGLE):
-                    sequence = 0
+                    sequence = Sequence(0)
             # create a TxIn object with the prev_tx, prev_index and sequence
             # the same as the current tx_in and the script_sig from above
             new_tx_in = TxIn(
@@ -336,7 +353,7 @@ class Tx:
             else:
                 s += tx_out.serialize()
         # add the locktime using int_to_little_endian in 4 bytes
-        s += int_to_little_endian(self.locktime, 4)
+        s += self.locktime.serialize()
         # add SIGHASH_ALL using int_to_little_endian in 4 bytes
         s += int_to_little_endian(hash_type, 4)
         # hash256 the serialization
@@ -352,7 +369,7 @@ class Tx:
                 all_prevouts += tx_in.prev_tx[::-1] + int_to_little_endian(
                     tx_in.prev_index, 4
                 )
-                all_sequence += int_to_little_endian(tx_in.sequence, 4)
+                all_sequence += tx_in.sequence.serialize()
             self._hash_prevouts = hash256(all_prevouts)
             self._hash_sequence = hash256(all_sequence)
         return self._hash_prevouts
@@ -416,14 +433,14 @@ class Tx:
         # add the value of the input in 8 bytes, little endian
         s += int_to_little_endian(tx_in.value(network=self.network), 8)
         # add the sequence of the input in 4 bytes, little endian
-        s += int_to_little_endian(tx_in.sequence, 4)
+        s += tx_in.sequence.serialize()
         # add the HashOutputs
         if (hash_type & 3) not in (SIGHASH_SINGLE, SIGHASH_NONE):
             s += self.hash_outputs()
         elif hash_type & SIGHASH_SINGLE == SIGHASH_SINGLE:
             s += self.tx_outs[input_index].serialize()
         # add the locktime in 4 bytes, little endian
-        s += int_to_little_endian(self.locktime, 4)
+        s += self.locktime.serialize()
         # add the sighash (SIGHASH_ALL) in 4 bytes, little endian
         s += int_to_little_endian(hash_type, 4)
         # hash256 the whole thing, interpret the as a big endian integer using int_to_big_endian
@@ -441,7 +458,7 @@ class Tx:
                 )
                 all_amounts += int_to_little_endian(tx_in.value(self.network), 8)
                 all_script_pubkeys += tx_in.script_pubkey(self.network).serialize()
-                all_sequence += int_to_little_endian(tx_in.sequence, 4)
+                all_sequence += tx_in.sequence.serialize()
             self._sha_prevouts = sha256(all_prevouts)
             self._sha_amounts = sha256(all_amounts)
             self._sha_script_pubkeys = sha256(all_script_pubkeys)
@@ -477,7 +494,7 @@ class Tx:
         s = b"\x00"
         s += int_to_byte(hash_type)
         s += int_to_little_endian(self.version, 4)
-        s += int_to_little_endian(self.locktime, 4)
+        s += self.locktime.serialize()
         if not hash_type & SIGHASH_ANYONECANPAY:
             s += self.sha_prevouts()
             s += self.sha_amounts()
@@ -493,7 +510,7 @@ class Tx:
             s += tx_in.prev_tx[::-1] + int_to_little_endian(tx_in.prev_index, 4)
             s += int_to_little_endian(tx_in.value(), 8)
             s += tx_in.script_pubkey().serialize()
-            s += int_to_little_endian(tx_in.sequence, 4)
+            s += tx_in.sequence.serialize()
         else:
             s += int_to_little_endian(input_index, 4)
         if hash_type & SIGHASH_SINGLE == SIGHASH_SINGLE:
@@ -562,7 +579,10 @@ class Tx:
 
     def verify(self):
         """Verify this transaction"""
-        if self.fee() < 0:
+        if self.fee() < self.vbytes():
+            print(
+                "This transaction won't relay without having a fee of at least {self.vbytes()}"
+            )
             return False
         for i in range(len(self.tx_ins)):
             if not self.verify_input(i):
@@ -613,21 +633,38 @@ class Tx:
         self.tx_ins[input_index].finalize_p2tr_keypath(sig)
         return self.verify_input(input_index)
 
-    def sign_p2tr_scriptpath(
-        self,
-        input_index,
-        private_key,
-        control_block,
-        tap_script,
-        hash_type=SIGHASH_DEFAULT,
-        aux=b"\x00" * 32,
-    ):
+    def initialize_p2tr_multisig(self, input_index, control_block, tap_script):
         tx_in = self.tx_ins[input_index]
-        tx_in.witness = Witness([tap_script.raw_serialize(), control_block.serialize()])
-        sig = self.get_sig_taproot(
-            input_index, private_key, ext_flag=1, hash_type=hash_type, aux=aux
-        )
-        tx_in.witness.items = [sig] + tx_in.witness.items
+        if len(tx_in.witness.items) == 0:
+            tx_in.witness = Witness(
+                [tap_script.raw_serialize(), control_block.serialize()]
+            )
+            if type(tap_script) != MultiSigTapScript:
+                raise RuntimeError("tap script must be MultiSigTapScript")
+            tx_in.tap_script = tap_script
+
+    def finalize_p2tr_multisig(self, input_index, sigs):
+        tx_in = self.tx_ins[input_index]
+        if len(tx_in.witness.items) < 2 or tx_in.tap_script is None:
+            raise RuntimeError("initialize single leaf multisig first")
+        for point in tx_in.tap_script.points:
+            for sig in sigs:
+                if len(sig) == 0:
+                    continue
+                elif len(sig) == 64:
+                    hash_type = SIGHASH_DEFAULT
+                    schnorr = SchnorrSignature.parse(sig)
+                elif len(sig) == 65:
+                    hash_type = sig[-1]
+                    schnorr = SchnorrSignature.parse(sig[:-1])
+                else:
+                    raise RuntimeError("invalid signature length")
+                msg = self.sig_hash(input_index, hash_type=hash_type)
+                if point.verify_schnorr(msg, schnorr):
+                    tx_in.witness.items.insert(0, sig)
+                    break
+            else:
+                tx_in.witness.items.insert(0, b"")
         return self.verify_input(input_index)
 
     def sign_input(
@@ -734,11 +771,10 @@ class Tx:
 
     def is_rbf_able(self):
         # https://github.com/bitcoin/bips/blob/master/bip-0125.mediawiki#Implementation_Details
-        eligible = False
         for tx_in in self.tx_ins:
-            if tx_in.sequence < 0xFFFFFFFF - 1:
-                eligible = True
-        return eligible
+            if tx_in.sequence.is_rbf_able():
+                return True
+        return False
 
     def find_utxos(self, address):
         """Returns transaction outputs that matches the address"""
@@ -761,23 +797,24 @@ class Tx:
 
 
 class TxIn:
-    def __init__(self, prev_tx, prev_index, script_sig=None, sequence=0xFFFFFFFF):
+    def __init__(self, prev_tx, prev_index, script_sig=None, sequence=None):
         self.prev_tx = prev_tx
         self.prev_index = prev_index
         if script_sig is None:
             self.script_sig = Script()
         else:
             self.script_sig = script_sig
-        self.sequence = sequence
+        if sequence is None:
+            self.sequence = Sequence()
+        else:
+            self.sequence = Sequence(sequence)
         self._value = None
         self._script_pubkey = None
         self.witness = Witness()
+        self.tap_script = None
 
     def __repr__(self):
-        return "{}:{}".format(
-            self.prev_tx.hex(),
-            self.prev_index,
-        )
+        return "{self.prev_tx.hex()}:{self.prev_index}"
 
     @classmethod
     def parse(cls, s):
@@ -793,7 +830,7 @@ class TxIn:
         # you can use Script.parse to get the actual script
         script_sig = Script.parse(s)
         # sequence is 4 bytes, little-endian, interpret as int
-        sequence = little_endian_to_int(s.read(4))
+        sequence = Sequence.parse(s)
         # return an instance of the class (cls(...))
         return cls(prev_tx, prev_index, script_sig, sequence)
 
@@ -806,7 +843,7 @@ class TxIn:
         # serialize the script_sig
         result += self.script_sig.serialize()
         # serialize sequence, 4 bytes, little endian
-        result += int_to_little_endian(self.sequence, 4)
+        result += self.sequence.serialize()
         return result
 
     def fetch_tx(self, network="mainnet"):
@@ -859,15 +896,15 @@ class TxIn:
 
     def finalize_p2wsh_multisig(self, signatures, witness_script):
         """Puts together the signatures for a p2wsh input so the input verifies."""
-        # the format for multisig is [b'\x00', each signature, then the WitnessScript (raw-serialization)]
-        items = [b"\x00", *signatures, witness_script.raw_serialize()]
+        # the format for multisig is [b'', each signature, then the WitnessScript (raw-serialization)]
+        items = [b"", *signatures, witness_script.raw_serialize()]
         # set the witness of the input to be these items
         self.witness = Witness(items)
 
     def finalize_p2sh_p2wsh_multisig(self, signatures, witness_script):
         """Puts together the signatures for a p2sh-p2wsh input so the input verifies."""
-        # the format for multisig is [b'\x00', each signature, then the WitnessScript (raw-serialization)]
-        items = [b"\x00", *signatures, witness_script.raw_serialize()]
+        # the format for multisig is [b'', each signature, then the WitnessScript (raw-serialization)]
+        items = [b"", *signatures, witness_script.raw_serialize()]
         # set the witness of the input to be these items
         self.witness = Witness(items)
         # the RedeemScript is the p2wsh ScriptPubKey of the WitnessScript
@@ -885,7 +922,7 @@ class TxOut:
         self.script_pubkey = script_pubkey
 
     def __repr__(self):
-        return "{}:{}".format(self.amount, self.script_pubkey)
+        return f"{self.amount}:{self.script_pubkey}"
 
     def serialize(self):
         """Returns the byte serialization of the transaction output"""
@@ -908,3 +945,38 @@ class TxOut:
         script_pubkey = ScriptPubKey.parse(s)
         # return an instance of the class (cls(...))
         return cls(amount, script_pubkey)
+
+    @classmethod
+    def to_address(cls, address, amount):
+        """Takes an address and an amount and makes a TxOut object"""
+        if address.startswith("bc1") or address.startswith("tb1"):
+            _, version, h = decode_bech32(address)
+            if version == 0:
+                if len(h) == 20:
+                    script_pubkey = P2WPKHScriptPubKey(h)
+                elif len(h) == 32:
+                    script_pubkey = P2WSHScriptPubKey(h)
+                else:
+                    raise ValueError(f"{address} is not a valid bech32 address")
+            elif version == 1:
+                if len(h) == 32:
+                    script_pubkey = P2TRScriptPubKey(h)
+                else:
+                    raise ValueError(f"{address} is not a valid bech32 address")
+            else:
+                raise ValueError(f"{address} is an unknown type of segwit address")
+        elif address[0] in ("3", "2"):
+            h = decode_base58(address)
+            if len(h) == 20:
+                script_pubkey = P2SHScriptPubKey(h)
+            else:
+                raise ValueError(f"{address} is not a valid base58 p2sh address")
+        elif address[0] in ("1", "m", "n"):
+            h = decode_base58(address)
+            if len(h) == 20:
+                script_pubkey = P2PKHScriptPubKey(h)
+            else:
+                raise ValueError(f"{address} is not a valid base58 p2pkh address")
+        else:
+            raise ValueError(f"{address} is an unknown or invalid type of address")
+        return cls(amount=amount, script_pubkey=script_pubkey)
